"""CDISC XPT files cleaner.
Import the modules to work with the XPTCleaner:
"""
import yaml
from importlib import resources
from configparser import ConfigParser

# Supress warnings ....
import warnings
warnings.filterwarnings("ignore")

# Version of xptcleaner package
__version__ = "1.0.0"

# Read yaml config file
with resources.path("xptcleaner", "config.yaml") as path:
    f = open(str(path))
    CONFIG = yaml.load(f, Loader=yaml.FullLoader)
    print(CONFIG)
    f.close()



"""Main entry point for the XPT Cleaner.
"""
# Standard library imports
import sys
from importlib import resources
# xptcleaner and module xptclean import
import xptcleaner
from xptcleaner import xptclean

def main() -> None:
    print("running gen_vacab")
    # if multiple files provided and the same Synonym exists in both files and map to different submission values,
    # the one in the second file will overwrite the mapping in the first file
    # so you want to pass the CDISC published file as the last file in the list
    #xptclean.gen_vocab(["C:/Project/src/R/sendigRPkg/SEND_Terminology_EXTENSIBLE.txt",
    #          "C:/Project/src/R/sendigRPkg/SEND Terminology_2021_12_17.txt"],
    #          "C:/Project/src/R/sendigRPkg/SEND3.1_test.json")

    #print("running standardize_file")
    #jsonfile = "C:/Project/src/R/sendigRPkg/SEND3.1_test.json"
    #rawXptFolder = "C:/BioCelerate/TDSStudies/43066/"
    #cleanXptFolder = "C:/BioCelerate/TDSStudiesClean/43066/"
    #xptclean.standardize_file(rawXptFolder, cleanXptFolder, jsonfile)

if __name__ == "__main__":
    main()



import chardet
import pandas
import pyreadstat

from xptcleaner.cdisc_mapping.logger import logger
from xptcleaner.cdisc_mapping.file_mapping import xpt_mapping

LOG = logger.get_logger(__name__)


def standardize_text(s):
    """
    Returns standardized text
    """
    import re
    # print("orginal term = ", s)
    try:
        standard_s = re.sub("[^0-9a-zA-Z]+", " ", str(s)).upper()
        return standard_s.strip()
    except Exception as e:
        LOG.error("ERROR exception caught:  " + str(e))
    return False


def column_mapping(columns_to_map, df, json_file):
    for column in columns_to_map.keys():
        try:
            #df[column] = df[column].map( "necrosys new" "NECROSYS"
            df[column] = df[column].map(
                lambda x: xpt_mapping.do_mapping(columns_to_map[column], x, json_file)
            )
        except Exception as e:
            LOG.error("ERROR exception caught:  " + str(e))

    return df


def read_XPT(file_path, file_name):
    with open(file_path + file_name, "rb") as f:
        result = chardet.detect(f.read())
    xpt_df, meta = pyreadstat.read_xport(file_path + file_name, metadataonly=False, encoding=result["encoding"])
    #xpt_df, meta = pyreadstat.read_xport(file_path + file_name, encoding='ascii', metadataonly=False)
    return xpt_df, meta


def merge_fix_cols(df_left, df_right, uniqueID):
    df_merged = pandas.merge(
        df_left, df_right, how="left", left_on=uniqueID, right_on=uniqueID
    )
    for col in df_merged:
        if col.endswith("_x"):
            df_merged.rename(columns=lambda col: col.rstrip("_x"), inplace=True)
        elif col.endswith("_y"):
            to_drop = [col for col in df_merged if col.endswith("_y")]
            df_merged.drop(to_drop, axis=1, inplace=True)
        else:
            LOG.error("No column conflicts during merge")

    return df_merged



import pandas
import os
import json

from xptcleaner.cdisc_mapping.logger import logger
from xptcleaner.cdisc_mapping.utils import utils
pandas.options.mode.chained_assignment = None

LOG = logger.get_logger(__name__)

def check_valid_files(input_dir):
    """
    Returns true if dm.xpt is available, otherwise false
    """
    input_xpt = os.listdir(input_dir)
    input_xpt = [i for i in input_xpt if i.endswith('.xpt')]
    input_xpt_cleaned = [x.lower() for x in input_xpt]
    files_dict = dict(zip(input_xpt_cleaned, input_xpt))
    #search in files_dict.keys() for cleaned file name
    required_xpts = ["dm.xpt"]
    optional_xpts = ["mi.xpt", "ds.xpt", "ex.xpt", "ts.xpt", "tx.xpt"]
    combo_xpts = [required_xpts, optional_xpts]
  
    # Get the XPT files that will not be mapped
    xpt_noMap = [v for v in list(files_dict.keys()) if v not in set().union(*combo_xpts)]

    xpt_noMap_dict = dict((d, files_dict.pop(d, None)) for d in xpt_noMap)
    required_dict = dict((d, files_dict.pop(d, None)) for d in required_xpts)

    # Logic depending if dm.xpt found or not
    if None not in required_dict.values(): #dm found
        return True, required_dict, files_dict, xpt_noMap_dict
    else: #dm not found
        return False, required_dict, files_dict, xpt_noMap_dict

def do_mapping(column, term, json_file):
    """
    Returns mapped terms from column and provided
    vocab file
    """

    with open(json_file, "r") as j:
        contents = json.loads(j.read())

    # make sure we have the column
    if column not in contents.keys():
        print("Column: " + column + "was not found in keys")
    else:
        # consider doing text standardization
        stan_term = utils.standardize_text(term)
        if stan_term in contents[column]:
            mapped_term = contents[column][stan_term]
            if (stan_term != mapped_term):
                LOG.info(
                    '"' + term + '","' + stan_term + '","' + mapped_term+ '", ' + column
                )
            return contents[column][stan_term]
        elif stan_term == '':
            # Possibly warn this is blank?
            return ""
        else:
            LOG.warning(
                '"' + term + '","' + stan_term + '",, ' + column
            )
            # leave the raw value as it is, it is up to the user to check the log file and correct all the mapping
            return term

def MI_dataframe(xpt_dir, xpt_dict, json_file):

    mi_xpt, meta  = utils.read_XPT(xpt_dir, xpt_dict['mi.xpt'])
    # General Histopathologic Exam, Qual
    dfMI= mi_xpt[mi_xpt["MITESTCD"] == "GHISTXQL"]
    if len(dfMI) == 0:
        #This may be a 3.0 dataset with MITESTCD = "MIEXAM"
        dfMI = mi_xpt[mi_xpt["MITESTCD"] == "MIEXAM"]
    if len(dfMI) != len(mi_xpt):
        MIdiff = len(mi_xpt) - len(dfMI)
        LOG.warning(
            str(MIdiff)
            + " rows are not 'General Histopathologic Exam, Qual' or 'Microscopic Examination' and will be excluded"
        )

    # Do mapping on available columns
    columns_to_map = {
        "MISTRESC": "Non-Neoplastic Finding Type",
        "MISPEC": "Specimen",
        # Anatomical Location is not for MIANTREG, it is for FXLOC
        # "MIANTREG": "Anatomical Location",
        "MISEV": "SEND Severity"
    }

    dfMI_map = utils.column_mapping(columns_to_map, dfMI, json_file)

    return dfMI_map, meta


def EX_dataframe(xpt_dir, xpt_dict, json_file):

    # EX.xpt takes priority for route over TS.xpt
    # TODO: need flexible name strategy
    ex_xpt, meta = utils.read_XPT(xpt_dir, xpt_dict['ex.xpt'])
    # print(ex_xpt)
    # Do mapping on available columns
    columns_to_map = {"EXROUTE": "Route of Administration Response"}
    dfEX_map = utils.column_mapping(
		columns_to_map, 
		ex_xpt, 
		json_file)
    return dfEX_map, meta


def TS_dataframe(xpt_dir, xpt_dict, json_file):
    TS_xpt, meta = utils.read_XPT(xpt_dir, xpt_dict['ts.xpt'])
    #Map the TS rows of interest
    columns_to_map = {"ROUTE": "Route of Administration Response",
                      "SPECIES": "Species",
                      "STRAIN": "Strain/Substrain"}

    for index, row in TS_xpt.iterrows():
        for col_short in row[["TSPARMCD"]]:
            col_long = columns_to_map.get(col_short)
            if (col_long is not None):
                TS_xpt.at[index, "TSVAL"]=do_mapping(col_long, TS_xpt.at[index, "TSVAL"], json_file)
    return TS_xpt, meta

def TX_dataframe(xpt_dir, xpt_dict, json_file):
    TX_xpt, meta = utils.read_XPT(xpt_dir, xpt_dict['tx.xpt'])
    #Map the TS rows of interest
    columns_to_map = {"TCNTRL": "SEND Control Type"}

    for index, row in TX_xpt.iterrows():
        for col_short in row[["TXPARMCD"]]:
            col_long = columns_to_map.get(col_short)
            if (col_long is not None):
                TX_xpt.at[index, "TXVAL"]=do_mapping(col_long, TX_xpt.at[index, "TXVAL"], json_file)
    return TX_xpt, meta

def DS_dataframe(xpt_dir, xpt_dict, json_file):

    ds_xpt, meta = utils.read_XPT(xpt_dir, xpt_dict['ds.xpt'])

    # Do mapping on available columns
    columns_to_map = {"DSDECOD": "Standardized Disposition Term"}
    dfDS_map = utils.column_mapping(
        columns_to_map, ds_xpt, json_file
    )

    return dfDS_map, meta


def DM_dataframe(xpt_dir, xpt_dict, json_file):

    dm_xpt, meta = utils.read_XPT(xpt_dir, xpt_dict['dm.xpt'])

    columns_to_map = {"SEX": "Sex"}
    dfDM_map = utils.column_mapping(
        columns_to_map, dm_xpt, json_file
    )

    return dfDM_map, meta



import logging
import sys

APP_LOGGER_NAME = "CDISC Mapper"


def setup_applevel_logger(logger_name=APP_LOGGER_NAME, is_debug=True, file_name=None):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG if is_debug else logging.INFO)

    formatter = logging.Formatter("%(asctime)s,%(levelname)s,%(message)s", "%Y-%m-%d %H:%M:%S")

    sh = logging.StreamHandler(sys.stdout)
    sh.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(sh)

    if file_name:
        fh = logging.FileHandler(file_name)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        # Add the first line as the header
        logger.info("Original Term,Standardized Term,Mapped Term,CodeList")

    return logger


def get_logger(module_name):
    return logging.getLogger(APP_LOGGER_NAME).getChild(module_name)



import pandas

from xptcleaner.cdisc_mapping.logger import logger
from xptcleaner.cdisc_mapping.utils import utils

LOG = logger.get_logger(__name__)


def gen_vocab(in_file, section_list):
    """
    Returns dict of dicts representing each section list
    synonym -> preferred term mapping from tab-separated
    term file.
    """

    result_dict = {}

    for section in section_list:
        # start extensible, finish CDISC to overwrite conflicts
        vocab_dict = {}
        sect_df = pandas.DataFrame()
        for f in in_file:
            # print(f)
            with open(f, 'r') as my_in_file:
                vocab_df = pandas.read_csv(my_in_file, sep="\t")

            # Filter dataframe to necessary section
            temp_df = vocab_df[['CDISC Submission Value', 'CDISC Synonym(s)']][vocab_df["Codelist Name"].isin([section])]
            temp_df['CDISC Submission Value'].str.upper()
            temp_df['CDISC Synonym(s)'].str.upper()
            sect_df = pandas.concat([sect_df, temp_df], ignore_index=True)

        # Go through the rows and add the data into the blank
        # vocab dict, replace missing values with empty string
        sect_df.apply(lambda row: make_dict(row.fillna(""), vocab_dict), axis=1)
        result_dict[section] = vocab_dict

    return result_dict

def make_dict(row, vocab_dict):
    """
        Returns diction of synonyms and preferred terms
        from a pandas row
    """

    synonyms = row['CDISC Synonym(s)'].split(';')

    # Standardize text
    synonyms = list(map(utils.standardize_text, synonyms))
    for synonym in filter(None, synonyms):
         vocab_dict[synonym] = row['CDISC Submission Value']

    # Include the preferred term as another lookup term
    standardized_perferred = utils.standardize_text(row['CDISC Submission Value'])
    vocab_dict[standardized_perferred] = row['CDISC Submission Value']

    return vocab_dict


def compare_vocabs(vocab_a, vocab_b):
    """
    Returns a dict of keys in vocab_a that aren't in b,
    and keys in b not in a
    """
    vocab_a_terms = []
    vocab_b_terms = []

    groups_a = vocab_a.keys()
    groups_b = vocab_b.keys()

    diff_group_avb = list(set(groups_a) - list(set(groups_b)))
    diff_group_bva = list(set(groups_b) - list(set(groups_a)))

    # This should be rare, but if the section columns are different
    # issue a warning -- these are normally specified in the
    # config

    if len(diff_group_avb) > 0 or len(diff_group_bva) > 0:
        LOG.warn("Section headings not matching!")

    # Check the terms themselves
    for group in groups_a:
        vocab_a_terms.extend(vocab_a[group].keys())
    for group in groups_b:
        vocab_b_terms.extend(vocab_b[group].keys())

    a_not_in_b = list(set(vocab_a_terms) - set(vocab_b_terms))
    b_not_in_a = list(set(vocab_b_terms) - set(vocab_a_terms))

    return a_not_in_b, b_not_in_a



"""Setup script for xptcleaner"""

# Standard library imports
import pathlib
import os

# Third party imports
from setuptools import setup
from setuptools import setup, find_namespace_packages, find_packages

# Utility function to read the README.md file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README.md file and 2) it's easier to type in the README.md file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

# The directory containing this file
HERE = pathlib.Path(__file__).resolve().parent

# The text of the README file is used as a description
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="xptcleaner",
    version="1.0.0",
    description="CDISC SEND xpt (SAS v5 Transport format) files cleaner",
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    url="https://github.com/phuse-org/sendigR/tree/main/python/xptcleaner",
    author="Cmsabbir Ahmed, Yousuf Ali, Susan Butler, Michael Denieu, William Houser, Brianna Paisley, Michael Rosentreter, Kevin Snyder, Wenxian Wang",
    author_email="cmsabbir.ahmed@fda.hhs.gov, md.ali@fda.hhs.gov, susan.butler@fda.hhs.gov, michael.denieu@labcorp.com, william.houser@bms.com, paisley_brianna_meadow@lilly.com, michael.rosentreter@bayer.com, kevin.snyder@fda.hhs.gov, wenxian.wang@bms.com",
    license=read("LICENSE"),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(include=['xptcleaner', 'xptcleaner.*', 
    'xptcleaner.cdisc_mapping.*','xptcleaner.cdisc_mapping.file_mapping.*', 
    'xptcleaner.cdisc_mapping.vocab_management.*', 
    'xptcleaner.cdisc_mapping.logger.*',
    'xptcleaner.cdisc_mapping.utils.*']),
    include_package_data=True,
    install_requires=["pandas", "pyreadstat"],
    entry_points={"console_scripts": ["xptceaner=xptceaner.__main__:main"]},
) 


# Supress warnings ....
import warnings
import os
import json
import shutil
import pyreadstat
from xptcleaner import CONFIG

from .cdisc_mapping.logger import logger
from .cdisc_mapping.file_mapping import xpt_mapping
from .cdisc_mapping.vocab_management import create_vocab

warnings.filterwarnings("ignore")

# Set up log file
logfile = CONFIG["log_file"]
print(logfile)
LOG = logger.setup_applevel_logger(file_name=str(logfile))

def gen_vocab(in_file, out_path):
    """
    Create json file for vocabulary mappings.
    Keys are synonyms and values are the CDISC Controlled Terminology Submission values.
    Vocabularies are defined by column values from the tab-delimited files.

    Parameters
    ----------
    in_file : str
        List of tab-delimited files with synonyms and preferred terms.
    out_path : str
        output json filename.

    """
    section_cols = CONFIG["section_cols"]

    vocab = create_vocab.gen_vocab(in_file, section_cols)

    with open(out_path, "w") as out_file:
        out_file.write(json.dumps(vocab, indent=4, sort_keys=True))


def compare_vocabs(vocab_path_a, vocab_path_b):
    """
    Reports if vocabs are same, and different terms in each
    """
    with open(vocab_path_a, "r") as json_file:
        vocab_a = json.loads(json_file.read(), strict=False)

    with open(vocab_path_b, "r") as json_file:
        vocab_b = json.load(json_file)

    a_not_in_b, b_not_in_a = create_vocab.compare_vocabs(vocab_a, vocab_b)

    if len(a_not_in_b) != 0 and len(b_not_in_a) != 0:
        LOG.warning(
            "Vocabs not matching:  vocab a not in b: "
            + str(len(a_not_in_b))
            + " vocab b not in a "
            + str(len(b_not_in_a))
        )

def standardize_file(input_xpt_dir, output_xpt_dir, json_file):
    """
    Standardizes SEND xpt files using CDISC controlled terminologies.
    Here is the list of CDISC codelist supported.
    - Sex
    - Strain/Substrain
    - Species
    - SEND Severity
    - Route of Administration Response
    - Standardized Disposition Term
    - Specimen
    - Non-Neoplastic Finding Type

     Parameters
    ----------
    input_xpt_dir : str
        input folder name with xpt files under the folder.
    output_xpt_dir : str
        output folder name for writing the cleaned xpt files.
    json_file : str
        json filename used for mapping.
    """
# Read in the file
    is_valid, dm_dict, optional_files_dict, xpt_extra = xpt_mapping.check_valid_files(input_xpt_dir)
    #create the output folder if not exist, remove the xpt files if exist in the folder
    isExist = os.path.exists(output_xpt_dir)
    if not isExist:
        os.makedirs(output_xpt_dir)
    for f in os.listdir(output_xpt_dir):
        if f.endswith(".xpt"):
            os.remove(os.path.join(output_xpt_dir, f))

    if is_valid:
        dfDM, meta = xpt_mapping.DM_dataframe(input_xpt_dir, dm_dict, json_file)
        pyreadstat.write_xport(dfDM, output_xpt_dir + "/" + "dm.xpt",
                               file_format_version=5,
                               table_name="DM",
                               file_label="Demographics",
                               column_labels=meta.column_labels
                               )

        if "ex.xpt" in optional_files_dict.keys():
            dfEX, meta = xpt_mapping.EX_dataframe(input_xpt_dir, optional_files_dict, json_file)
            pyreadstat.write_xport(dfEX, output_xpt_dir + "/" + "ex.xpt",
                                   file_format_version=5,
                                   table_name="EX",
                                   file_label="Exposure",
                                   column_labels=meta.column_labels
                                   )

        if "ts.xpt" in optional_files_dict.keys():
            dfTS, meta = xpt_mapping.TS_dataframe(input_xpt_dir, optional_files_dict, json_file)
            pyreadstat.write_xport(dfTS, output_xpt_dir + "/" + "ts.xpt",
                                   file_format_version=5,
                                   table_name="TS",
                                   file_label="Trial Summary",
                                   column_labels=meta.column_labels
                                   )

        if "tx.xpt" in optional_files_dict.keys():
            dfTx, meta = xpt_mapping.TX_dataframe(input_xpt_dir, optional_files_dict, json_file)
            pyreadstat.write_xport(dfTx, output_xpt_dir + "/" + "tx.xpt",
                                   file_format_version=5,
                                   table_name="TX",
                                   file_label="Trial Set",
                                   column_labels=meta.column_labels
                                   )

        if "ds.xpt" in optional_files_dict.keys():
            dfDS, meta = xpt_mapping.DS_dataframe(input_xpt_dir, optional_files_dict, json_file)
            pyreadstat.write_xport(dfDS, output_xpt_dir + "/" + "ds.xpt",
                                   file_format_version=5,
                                   table_name="DS",
                                   file_label="Disposition",
                                   column_labels= meta.column_labels
                                   )

        if "mi.xpt" in optional_files_dict.keys():
            dfMI, meta = xpt_mapping.MI_dataframe(input_xpt_dir, optional_files_dict, json_file)
            pyreadstat.write_xport(dfMI, output_xpt_dir + "/" + "mi.xpt",
                                   file_format_version=5,
                                   table_name="MI",
                                   file_label="Microscopic Findings",
                                   column_labels=meta.column_labels
                                   )
        for xpt in xpt_extra:
            shutil.copy(input_xpt_dir + "/" + xpt, output_xpt_dir + "/" + xpt)

    else:
        LOG.error("Missing required xpt files")
